/* Webroute 1.3.0
 * Simple WEB Proxy for Linux and perhaps other systems
 *     by Willy Tarreau
 *
 * webroute    This program is used to redirect HTTP requests
 *             it receives to the right HTTP server, or to
 *             another instance of itself, on an other host.
 *	       It acts like a proxy and all the Web browsers
 *	       that will have to use it must be setup to use
 *	       it as the HTTP Proxy. It then allows several
 *	       hosts on a network to access the Web via one
 *	       only server, which is particularly interesting
 *	       in case of a server connected to an Internet
 *	       provider via a modem with PPP.
 *
 *             The second feature of this program is to 
 *             automagically connect and/or disconnect a modem
 *	       to/from the line to allow users to connect to a
 *	       provider without having to do anything. It's
 *	       then possible to make the first connection
 *	       immediately after the first request arrives, and
 *	       to release it a few minutes after. Obviously,
 *	       all requests that arrive while the connection is
 *	       on can benefit of it, and won't have to connect
 *     	       it again.
 *
 *             One interesting aspect is that it doesn't require
 *             superuser privileges to run  :-)
 *
 * $Id: webroute.c,v 1.3.0 1997/05/28 17:04:13 willy Exp $
 *
 * Authors:
 *            - core by Willy Tarreau <tarreau@aemiaif.ibp.fr>
 *            - POST method added by Jay Evans <jevans@ionet.net>
 *              (Thanks to him for his support !)
 *
 * Changes:
 *   1.0.0 -> 1.0.1
 *             - Corrected a little bug which could be source
 *               of zombies at high loads
 *   1.0.1 -> 1.1.0
 *             - automatic kill when no request come during a
 *                predefined time
 *             - a way to allow or refuse connections from
 *               certain domains/hosts that the user can
 *               specify in a file
 *             - but no man page yet :-(
 *
 *   1.1.0 -> 1.2.0 ( upgrade by Jay Evans <jevans@ionet.net> )
 *             - POST method is now supported
 *             - Corrects a bug in MSIE 3.0 in the 'Content-Length'
 *               field syntax
 *
 *   1.2.0 -> 1.2.1
 *             - Corrected a bug after strncpy which didn't fill the address correctly.
 *               Sometimes, one may have got messages like 'server error, bad address'.
 *
 *   1.2.1 -> 1.2.2
 *             - Added a condition after accept to prevent webroute from
 *               getting down when receiving a non blocked signal.
 *             - Modified the error syntax to make sure that the error page
 *               won't be cached.
 *             - Added a #define for Sparc 1/2/4/10/ELC/IPX and perhaps other Sparcs
 *               under SunOS 4.1.X
 *               Thanks to Sander Chandon Janssens (S.Chandon@fokkerspace.nl) for this.
 *
 *   1.2.2 -> 1.2.3
 *             - Better support for "If-Modified-Since" and "Pragma: no-cache" headers
 *               to allow a "not modified" reply without connecting to the remote site
 *
 *   1.2.3 -> 1.2.4
 *             - Corrected a bug introduced in 1.2.3 that made the connection script
 *               be launched sometimes, even if the requested page was cached.
 *             - Displays server PID so that it can be easy to use it in scripts.
 *
 *   1.2.4 -> 1.3.0
 *             - SOCKS support added by Sylvan Butler <sylvan@cyberhighway.net>
 *             - slightly improved Makefile
 *
 * Todo:       - a man page or even several man pages :)
 *               perhaps one day ...
 *             - add an option to supply an access log with
 *               hostnames and requests.
 *             - support compression between 2 webroute proxies
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>

#ifdef SPARC
#include <sys/socket.h>
#endif

#ifdef SOCKS
#define connect Rconnect
#define getsockname Rgetsockname
#define bind Rbind
#define accept Raccept
#define listen Rlisten
#define select Rselect
#endif

/* Let's define Webroute's release as VER.REL.BUG */
#define WR_VER 1
#define WR_REL 3
#define WR_BUG 0

/* default listen port */
#define LISTENPORT  8080

/* default timeout for any read or write, in seconds */
#define TIMEOUT     120

/* length of data buffer in chars */
#define LDATA       1024

/* length of remote server address */
#define LADR        128

/* default port to connect to if unspecified */
#define DEFAULTPORT 80

/* max length of access file */
#define MAXACCESS 100

/* Max header lines in the request */
#define MAXHDRS   16

/* correct something Sun may have forgotten :-) */
#ifdef SPARC
extern char *optarg;
#endif

int  ConnectToProxy = 0;               /* 1 here means this program will connect to another instance of it
                                          0 means we'll connect directly to the Internet */
char NextProxyAdr[128];                /* the name of the host where the next instance of the program runs */
int  NextProxyPort;                    /* and its port */
int  NoCache=0;                        /* if not 0, prevents web browsers from retrieving pages in their own
					  cache when the users does a "Reload" action */
int  NeverMod=0;                       /* if not 0, when the proxy is asked for an "if-modified" page, it always
					  replies that it has not been modified, to force to use local cache.*/

char connScript[100];                  /* name of the Opening Connection Script. (for modem only) */
char disconScript[100];                /* name of the Closing Connection Script. (for modem only) */
char accessfile[100];                  /* name of the access list file */
int  killtimeout=0,                    /* seconds between last connection and killing */
    conntimeout=0;                    /* seconds between end of reading and closing connection   */
int  Connected=0;                      /* 1 if connection is useable, 0 if not  (for modem only)  */

struct sockaddr_in OutputSocket;       /* output socket descriptor */
int sockOut;                           /* the socket itself */

int listenport=LISTENPORT;
struct sockaddr_in ListenSocket, InputSocket;
int sockListen,sockIn,lgInputSocket;
int val=1;                             /* used many times by setsockopt */
int childnbr=0;                        /* counts how many childs we have */
int webroutepid;

typedef struct {
    char host[128];
    char grant;
} access_s;

int use_access=0;                      /* by default, no access list */
int accesslen=0;
access_s accesslist[100];

char  *BADREQ=
"HTTP/1.0 400 Parse Error\n"
"Server: webroute\n"
"Expires: Thu, 01 Jan 1970 00:00:00 GMT\n"
"Content-Type: text/html\n\n"

"<HEAD><TITLE>ERROR: Request is not valid</TITLE></HEAD>\n"
"<BODY><H1>ERROR</H1><H2>Request is not valid</H2><HR><P>"
"Webroute release %d.%d.%d running on server %s:<BR>"
"Your browser sent a query that this server could not"
"understand.<P>"
"Reason: Invalid or unsupported method.<HR><P>";


char  *SERVERR=
"HTTP/1.0 500 Server Error\n"
"Server: webroute\n"
"Expires: Thu, 01 Jan 1970 00:00:00 GMT\n"
"Content-Type: text/html\n\n"

"<HEAD><TITLE>ERROR: Server encountered an error</TITLE></HEAD>\n"
"<BODY><H1>ERROR</H1><H2>Server encountered an error</H1><HR><P>"
"Webroute release %d.%d.%d running on server %s:<BR>"
"Internal proxy error while processing your query.<P>"
"Reason: Invalid query or bad address.<HR><P>";

char  *CONNERR=
"HTTP/1.0 500 Server Error\n"
"Server: webroute\n"
"Expires: Thu, 01 Jan 1970 00:00:00 GMT\n"
"Content-Type: text/html\n\n"

"<HEAD><TITLE>ERROR: Network link couldn't be set up</TITLE></HEAD>\n"
"<BODY><H1>ERROR</H1><H2>Network link couldn't be set up</H2><HR><P>"
"Webroute release %d.%d.%d running on server %s:<BR>"
"Failed to launch opening connection script.<P>"
"Reason: The Modem line may be busy.<HR><P>";

char  *ABORTOUTERR=
"HTTP/1.0 500 Server Error\n"
"Server: webroute\n"
"Expires: Thu, 01 Jan 1970 00:00:00 GMT\n"
"Content-Type: text/html\n\n"

"<HEAD><TITLE>ERROR: Connection has aborted</TITLE></HEAD>\n"
"<BODY><H1>ERROR</H1><H2>Connection has aborted</H2><HR><P>"
"Webroute release %d.%d.%d running on server %s:<BR>"
"The outgoing connection has just aborted. I can't go further.<P>"
"Reason: External line may be too slow or overloaded.<HR><P>";

char  *ABORTINERR=
"HTTP/1.0 500 Server Error\n"
"Server: webroute\n"
"Expires: Thu, 01 Jan 1970 00:00:00 GMT\n"
"Content-Type: text/html\n\n"

"<HEAD><TITLE>ERROR: It seems as I can't send you anything more</TITLE></HEAD>\n"
"<BODY><H1>ERROR</H1><H2>It seems as I can't send you anything more</H2><HR><P>"
"Webroute release %d.%d.%d running on server %s:<BR>"
"Your connection seems to have been broken. I can't go further.<P>"
"Reason: You may have tried to stop, or your connection to Webroute may be too slow.<HR><P>"
"Note: logically, you should never see this message, so in this case please post me a mail !<BR>";

char  *ACCESSERR=
"HTTP/1.0 403 Access Denied\n"
"Server: webroute\n"
"Expires: Thu, 01 Jan 1970 00:00:00 GMT\n"
"Content-Type: text/html\n\n"
"<HEAD><TITLE>ERROR: You have no permission to use Webroute on this server</TITLE></HEAD>\n"
"<BODY><H1>ERROR</H1><H2>You have no permission to use Webroute on this server</H2><HR><P>"
"Webroute release %d.%d.%d running on server %s:<BR>"
"Your host isn't allowed to use Webroute from this server. Please contact this server's administrator"
" or the Webroute maintainer (see Webroute's access file).<HR><P>";

char  *MSGNOTMOD=
"HTTP/1.0 304 Not modified\n"
"Server: webroute\n\n";

int connectto(char *hote, int port) {
    struct hostent *ServerName;

    /*   fprintf(stderr,"connectto: %s:%d\n",hote,port);*/
    if ((sockOut=socket(AF_INET, SOCK_STREAM, 0))==-1) {
	return(1);
    }
    if ((ServerName=gethostbyname(hote))==NULL) {
	return 1;
    }
    bzero((char *)&OutputSocket, sizeof(OutputSocket));
    OutputSocket.sin_family= AF_INET;
    OutputSocket.sin_port= htons(port);
    bcopy((char *)ServerName->h_addr,
	  (char *)&OutputSocket.sin_addr,
	  ServerName->h_length);

    if (connect(sockOut,(struct sockaddr*)&OutputSocket,sizeof(OutputSocket))==-1) {
	close(sockOut);
	return 1;
    }
    return 0;   /* connection OK */
}

/* this function should be called only by the child */
void sayerror(char *msg) {
    char errhost[500];
    char host[128];
    char *bottom = 
	"<H3>You can send bug reports to me: <A HREF=\"mailto:tarreau@aemiaif.ibp.fr\">Willy Tarreau</A></H3>"
	"<BR><H3>You can also check if there is a new release available at this moment:<BR><UL><LI>"
	"<A HREF=\"http://www-miaif.ibp.fr/willy/webroute/\">http://www-miaif.ibp.fr/willy/webroute/<LI>"
	"<A HREF=\"ftp://aemiaif.ibp.fr/pub/webroute/webroute/\">ftp://aemiaif.ibp.fr/pub/webroute/</A></UL></H3>"
	"</BODY>\n";

    gethostname(host,127);host[127]=0;
    sprintf(errhost,msg,WR_VER,WR_REL,WR_BUG,host);
    write(sockIn,errhost,strlen(errhost));
    write(sockIn,bottom,strlen(bottom));
    close(sockOut);
    close(sockIn);
    exit(1);   /* kill the child */
}

/* timeout alarm. Called from the child only */
void alarme(int sig) {
    close(sockOut);
    close(sockIn);
    close(sockListen);
    exit(1);  /* kill the child */
}

/* kill timeout alarm. Called from the waiter only */
void autokill(int sig) {
    close(sockOut);
    close(sockIn);
    close(sockListen);
    kill(getpid(),SIGHUP);  /* kill the waiter, and then the children */
    exit(0); /* in case of no success :-) */
}


/* The 2 following functions call a script. It's necessary to
   disable the SIGCHLD signal because if we don't, it will be
   generated after the script ends. That's not a problem to
   disable this signal because when the scripts are called,
   this means no connection is still active.
*/

/* Close modem connection if timeout has expired */
void autodiscon(int sig) {
    int err=0;

    signal(SIGUSR2,autodiscon);
    signal(SIGALRM,SIG_IGN); /* disables automatic disconnection */
    if (Connected) {
	if (*disconScript) {
	    signal(SIGCHLD,SIG_IGN); /* prevents this signal from calling */
	    err=system(disconScript);  /* launches disconnection script */
	}
    }
    switch (err) {
    case 0:              /* Disconnect OK */
	Connected=0;
	break;
    case 1:              /* the link was already disconnected */
	Connected=0;
	break;
    case 2:              /* DO NOT DISCONNECT YET, PLEASE ! (perhaps manual use ?)*/
	Connected=1;
	break;
    case 3:              /* Disconnection failed. Let's try again ! */
    default:             /* or any case we can't imagine */
	Connected=1;       /* still connected !!! */
	alarm(conntimeout);   /* retry after the same time-out */
	signal(SIGALRM,autodiscon);
	break;
    }
}

/* Open modem connection if needed */
int initautoconn() {
    int err=0;

    signal(SIGALRM,SIG_IGN); /* disables automatic disconnection */
    if (!Connected) {
	if (*connScript) {
	    signal(SIGCHLD,SIG_IGN);   /* prevents this signal from calling */
	    err=system(connScript);   /* launches connection script */
	}
    }
    if (!err)
	kill(webroutepid,SIGUSR1);
    return err;
}


/* allows 'Connected' to be set from any child */
void sigusr1(int sig) {
    signal(SIGUSR1, sigusr1);
    Connected=1;
}

/* This function clears zombies. It's used only by the parent process because it tells when the
   transfert ends, so it is possible to initialize the timeout for disconnection.
*/
void sigchld(int sig) {
    signal(SIGUSR2,autodiscon);
    signal(SIGCHLD,sigchld);
    wait(NULL);
    
    if (((--childnbr)<=0) && (conntimeout)) {  /* it was the last child so we start decounting */
	signal(SIGALRM, SIG_IGN);  /* before disconnection if timeout enabled */
	alarm(conntimeout);
	signal(SIGALRM,autodiscon);
    }
}



/* displays the right syntax to call Webroute */
void displaysyntax() {
    fprintf(stderr,"Syntax:\n");
    fprintf(stderr,"webroute [ -p port ] [ -x h:p ] [ -c con ] [ -d discon ] [ -{t|k} timeout ] [ -n ]\n");
    fprintf(stderr,"         [ -a access_file ] [-m]\n");
    fprintf(stderr,"Available options are:\n");
    fprintf(stderr,"  -p allows you to run webroute on the port <port>.\n");
    fprintf(stderr,"     If you don't have superuser privileges, you must use a port > 5000, since\n");
    fprintf(stderr,"     ports 1024 to 4999 could be in use for external usage by whatever process.\n");
    fprintf(stderr,"     The default port is %d.\n",LISTENPORT);
    fprintf(stderr,"  -x enables multi-proxy feature. This means that this instance of Webroute\n");
    fprintf(stderr,"     doesn't have itself access to the internet, but the one which is running\n");
    fprintf(stderr,"     on port <p> of host <h> can provide an access. It's possible to chain\n");
    fprintf(stderr,"     as many instances of Webroute as you want. That depends of your network\n");
    fprintf(stderr,"     topology\n");
    fprintf(stderr,"  -c means that when we have to re-establish a modem connection, we'll use the\n");
    fprintf(stderr,"     program or shell script which path if given as <con>\n");
    fprintf(stderr,"     WARNING: if you use -c but not -t, the modem will never be disconnected !\n");
    fprintf(stderr,"  -d means that when we have to disconnect the modem, we'll use the\n");
    fprintf(stderr,"     program or shell script which path if given as <discon>.\n");
    fprintf(stderr,"     This program will only be called after <timeout> seconds of inactivity.\n");
    fprintf(stderr,"  -t <timeout> specifies how many seconds the modem will stay connected when\n");
    fprintf(stderr,"     no request arrives. After this time, the <discon> program will be called.\n");
    fprintf(stderr,"     Can't be used if -k specified.\n");
    fprintf(stderr,"  -n prevents browsers from retrieving web pages from their own cache when the\n");
    fprintf(stderr,"     user asks for a \"Reload\". The page will then always be reloaded.\n");
    fprintf(stderr,"  -k <timeout> specifies how many seconds the webroute should wait for a\n");
    fprintf(stderr,"     connection before automatically killing itself. A 0 time means never\n");
    fprintf(stderr,"     kill (default). Can't be used if -t specified.\n");
    fprintf(stderr,"  -a <access_file> specifies a file which contains a list of hosts and\n");
    fprintf(stderr,"     domains which are authorized or not to connect to webroute.\n");
    fprintf(stderr,"  -m always tell the client that the requested page has not been modified\n");
    fprintf(stderr,"     so it can retrieve it from its own cache (speed up on slow links).\n");
    fprintf(stderr,"New releases available at:  http://www-miaif.ibp.fr/willy/webroute/\n");
}

void main(int argc, char **argv) {
    int opt;
    int contlen,getflag;

    webroutepid=getpid();
    contlen=getflag=0;
    printf("Webroute v %d.%d.%d - (C) 1996-97 Willy Tarreau <tarreau@aemiaif.ibp.fr>\n",WR_VER,WR_REL,WR_BUG);
    while ((opt=getopt(argc,argv,"p:x:c:d:t:nmk:a:"))!=-1) {
	switch (opt) {
	case ':' :  /* missing parameter */
	case '?' :  /* unknown option */
	    displaysyntax();
	    exit(1);

	case 'p' : /* port */
	    listenport=atoi(optarg);
	    break;

	case 'x' : {  /* external proxy */
	    char *p;
	    p=strchr(optarg,':');
	    if (p==NULL) {    /* unspecified port number. let's quit */
		fprintf(stderr,"missing port for next proxy\n");
		displaysyntax();
		exit(1);
	    }
	    *(p++)=0;  /* ends hostname */
	    NextProxyPort=atoi(p);
	    strcpy(NextProxyAdr,optarg);
	    ConnectToProxy=1;
	    break;
	}

	case 'c' : /* connection script name */
	    strcpy(connScript,optarg);
	    break;

	case 'd' : /* disconnection script name */
	    strcpy(disconScript,optarg);
	    break;

	case 't' : /* disconnect time-out */
	    conntimeout=atoi(optarg);
	    if (killtimeout) {
		fprintf(stderr,"Warning: autokill disabled because -t specified\n");
		killtimeout=0;
	    }
	    break;

	case 'k' : /* autokill time-out */
	    killtimeout=atoi(optarg);
	    if (conntimeout) {
		fprintf(stderr,"Warning: auto disconnect disabled because -k specified\n");
		conntimeout=0;
	    }
	    break;

	case 'a' : /* access file */
	    strcpy(accessfile,optarg);
	    use_access=1;
	    break;

	case 'n' : /* no cache */
	    NoCache=1;
	    break;

	case 'm' : /* never modified */
	    NeverMod=1;
	    break;
	} /* end of switch() */
    } /* end of while() */

#ifdef SOCKS
    SOCKSinit(argv[0]);
#endif

    if (use_access) {
	FILE *accf;
	char line[200];
	char *p;

	if ((accf=fopen(accessfile,"r"))==NULL) {
	    perror("Trying to open access file");
	    exit(1);
	}
	while ((accesslen < MAXACCESS) && (fgets(line,199,accf)!=NULL)){
	    if (line[0]!='#') { /* don't memorize comments */
		line[199]=0;  /* force end of line if line too big */
		while (p=strchr(line,'\t'))
		    *p=' ';
		if (sscanf(line,"%s %c\n",accesslist[accesslen].host,&accesslist[accesslen].grant)!=2) {
		    fprintf(stderr,"Warning: incorrect syntax for access list entry #%d: %s\n",accesslen+1,line);
		    fprintf(stderr,"The right syntax is:  host.domain    Y/N\n");
		}
		else {
		    accesslist[accesslen].grant=tolower(accesslist[accesslen].grant);
		    accesslen++;
		}
	    }
	}
	if (accesslen>=MAXACCESS)
	    fprintf(stderr,"Warning: Access list truncated to the %d first entries.\n",accesslen);
	else
	    fprintf(stderr,"%d patterns loaded in access list.\n",accesslen);
	fclose(accf);
    }

    /* initialization of listen socket */

    if ((sockListen=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))==-1) {
	perror("Webroute(socket)");
	exit(1);
    }

    if ((setsockopt(sockListen,SOL_SOCKET,SO_REUSEADDR,&val,sizeof(val))) == -1) {
	perror("Webroute(setsockopt)");
	exit(1);
    }
  
    bzero((char *)&ListenSocket, sizeof(ListenSocket));
    ListenSocket.sin_family=AF_INET;
    ListenSocket.sin_addr.s_addr=htonl(INADDR_ANY);
    ListenSocket.sin_port=htons(listenport);
    if (bind(sockListen,(struct sockaddr *)&ListenSocket,sizeof(ListenSocket))==-1) {
	perror("Webroute(bind)");
	exit(1);
    }
    if (listen(sockListen,5)==-1) {
	perror("Webroute(listen)");
	exit(1);
    }

    /* the socket is ready. Let's wait for requests ... */
    switch (webroutepid=fork()) {
    case -1:
	perror("Webroute(fork) failed while going daemon");
	exit(1);

    case 0 :  /* child becomes the daemon */
	/* let's close stdin, stdout, stderr to prevent messages from appearing on the console */
	close(0);
	close(1);
	close(2);
	webroutepid=getpid();  /* updates the pid */
	for (;;) {  /* infinite loop */
	    signal(SIGUSR1,sigusr1);
	    signal(SIGUSR2,autodiscon);
	    if (killtimeout) {
		signal(SIGALRM,autokill);
		alarm(killtimeout);  /* assumes auto-killing */
	    }

	    lgInputSocket=sizeof(InputSocket);

	    do {
		sockIn=accept(sockListen,(struct sockaddr *)&InputSocket,&lgInputSocket);
	    } while (sockIn==-1 && (errno==EINTR || errno==EAGAIN));   /* retries if interrupted */
	    if (sockIn==-1) {   /* if there's an error, we exit */
		exit(1);  /* don't wait when daemon is going sick ! */
	    }

	    /*      if (initautoconn()) {  /* launches automatic connection */
	    /*	write(sockIn,CONNERR,strlen(CONNERR));
		/*	close(sockIn);
			/*      }
				/*      else*/
	    switch (fork()) {
	    case -1 :
		close(sockIn);
		break;
		
	    case 0  : { /* a separate process for this request */  
		char data[LDATA];
		char adr[LADR],*p,*hdr;
		int ldata,lreq,port;
		int pragmanocache=0;  /* if 1, don't use cache. */
		int ifmod=0;    /* 1= "if-modified-since..." encountered */
		char headers[MAXHDRS][LDATA];  /* header lines in the request */
		int nbhdr=0;
		int direct=0;     /* direct copying when parameters are passed to GET */
		int err,i,reqcomplete=0;
		FILE *fsin;
		

		/* close listening socket because we don't need it yet */
		close(sockListen);

		/* let's reopen input socket as a file */
		if ((fsin=fdopen(sockIn,"rw"))==NULL)
		    sayerror(SERVERR);

		/* prepares for connection abort */

		signal(SIGALRM,alarme);

		/* avoid some sockets problems ... */
		val=1;
		setsockopt(sockIn,SOL_SOCKET,SO_LINGER,&val,sizeof(val));
		setsockopt(sockIn,SOL_SOCKET,SO_KEEPALIVE,&val,sizeof(val));
		      
		if (use_access) {
		    /* access verification 
		       The file structure is:
		       # comment
		       *               y|n
		       .               y|n
		       host.           y|n
		       host.domain     y|n
		       .domain         y|n

		       The analysis is going down the lines so the last matching line is taken for the authorization.
		       The advantage of such a method is that it becomes easy to include certain domains but some hosts
		       on that domain.

		       The dot ('.') designates the local network because no domain names are given on a local net.
		       So, 'pc7.' stands for 'pc7.my.domain'.
		       The star ('*') stands for everything. It's usefull to set the default access to 'Y' or 'N'.
		       It should be at the first line because it matches everything, so the previously analysed lines
		       would be ignored.
		    */
	    
		    struct sockaddr_in sa;
		    int sl=sizeof(sa);
		    struct hostent *h;
		    char adrascii[128];
		    int ent,l1,l2,auth;

		    if ((getpeername(sockIn,&sa,&sl)==-1) ||
			((h=gethostbyaddr(&sa.sin_addr,sizeof(struct in_addr),AF_INET))==NULL))
			sayerror(ACCESSERR);
		    /* le nom de la machine est dans h->h_name */
		    strcpy(adrascii,h->h_name);
		    if (!strchr(adrascii,'.'))  /* local network */ 
			strcat(adrascii,".");   /* let's add a '.' to the local name */
	     
		    l1=strlen(adrascii);
		    auth=0;  /* no authorization by default */
		    for (ent=0;ent<accesslen;ent++) {
			l2=strlen(accesslist[ent].host);
			if ((l1>=l2) && (!strncasecmp(&adrascii[l1-l2],accesslist[ent].host,l2)) ||
			    (!strcmp(accesslist[ent].host,"*"))) {
			    /* an inclusive entry has been found */
			    auth=accesslist[ent].grant=='y';
			}
		    }
		    /* here, if auth has gone down, access for this host is denied */
		    if (!auth)
			sayerror(ACCESSERR);
		}
	    

		/* here, we'll analyze the request and get rid of "http://adr:port". The address and port will
		   be duplicated and used to open the connection. */



		contlen=0;
		alarm(TIMEOUT);
		if (fgets(hdr=headers[0],LDATA,fsin)==NULL)
		    sayerror(BADREQ);
		nbhdr=1;
		if (!(direct=(strchr(headers[0],'?')!=NULL))) {
		    /* no parameter passed to GET, we don't have to connect directly */
		    while (nbhdr<MAXHDRS) {
			/*write(sockIn,hdr,strlen(hdr));		 */
			alarm(TIMEOUT);
			if (fgets(hdr=headers[nbhdr],LDATA,fsin)==NULL)
			    sayerror(SERVERR);
			ldata=strlen(hdr);
			if (!ldata || *hdr=='\n' || *hdr=='\r') { /* last header line */
			    reqcomplete=1;
			    break;
			}
			/* let's parse the request */
			if (!strncasecmp(hdr,"Content-length: ",16)){
			    contlen=atoi(hdr+16);
			}
			if (!strncmp(hdr,"If-Modified-Since",17))
			    if (!NoCache)
				ifmod=1;
			    else
				continue;  /* if "nocache", forget this line */
			if (!strncmp(hdr,"Pragma: no-cache",16))
			    if (!NoCache)
				pragmanocache=1;
			    else
				continue;  /* if "nocache", forget this line */
			nbhdr++;
		    }
	      
		    /* sprintf(data,"ifmod=%d, nev=%d, prg=%d, noc=%d\n",ifmod,NeverMod,pragmanocache,NoCache);
		       write(sockIn,data, strlen(data));*/

		    if (ifmod && NeverMod && !pragmanocache && !NoCache) { /* can reply immediately */
			do {
			    alarm(TIMEOUT);
			    err=write(sockIn,MSGNOTMOD,strlen(MSGNOTMOD));
			} while (err==-1 && errno==EINTR);
			if (err==-1)
			    sayerror(ABORTINERR);
			close(sockIn);   /* close the sockets */
			close(sockOut);
			exit(0);  /* no error */
		    }

		}

 		if (initautoconn())  /* launches automatic connection */
		    sayerror(CONNERR);

		strncpy(data,headers[0],LDATA);
		ldata=strlen(data);

		if (!ConnectToProxy)  {  /* if proxy-to-proxy connection, we don't modify the request */
		    char *str;


		    /* it's easy to log all requests here */
		    /* fprintf(stderr,"requete recue: %s",data);   */

		    if (strncmp(data,"GET ",4)==0) { /* we expect GET to be there */
			str=data+4;
			getflag=1;
		    } 
		    else if (strncmp(data,"POST ",5)==0) { /* we expect POST to be there */
			str=data+5;
			getflag=0;
		    }
		    else
			sayerror(BADREQ);

		    while (*str==' ')
			str++;

		    if (!strncmp(str,"http://",7))
			str+=7;

		    if ((p=strchr(str,'/'))!=NULL) {
			strncpy(adr,str,(p-str)); /* copies address in adr */
			adr[(p-str)]=0;  /* fills out the end of the address */
			str=p;  /* points to the rest of the request (without address) */
			lreq=ldata-(str-data);
		    }
		    else
			sayerror(BADREQ); /* if no /, error */
	        
		    /* at this stage, adr contains addr[:port], and str points to the local URL with the first  '/' */

		    if (adr[0]==0)
			sayerror(BADREQ);
	   
		    p=strchr(adr,':');
		    if (p==NULL) {      /* unspecified port. The default one will be used */
			port=DEFAULTPORT;
		    } else {            /* port is available. let's read it */
			*(p++)=0;       /* ends hostname */
			port=atoi(p);
		    }
		    /* end of request analysis. The hostname is in "adr", and the port in "port" */

		    if (connectto(adr,port))
			sayerror(SERVERR);
		
		    /* As it becomes a local URL, we only say "GET" and the end of the request. */
		    alarm(TIMEOUT);
		
		    if (getflag)
			write(sockOut,"GET ",4);
		    else
			write(sockOut,"POST ",5);
		
		    write(sockOut,str,lreq);
		}
		else { /* proxy-to-proxy connection ! */
		    if (connectto(NextProxyAdr,NextProxyPort))
			sayerror(SERVERR);
		    alarm(TIMEOUT);
		    write(sockOut,data,ldata);  /* write the normal request */
		}
	    
		if (!direct) {  /* let's copy the buffered headers */
		    for (i=1;i<nbhdr;i++) {
			alarm(TIMEOUT);
			write(sockOut,headers[i],strlen(headers[i]));
		    }
		}
	    
		/* now, let's copy all what we don't have copied yet */
		if (!reqcomplete) {
		    while (1) {
			alarm(TIMEOUT);
			if (fgets(data,LDATA,fsin)==NULL)
			    break;
			ldata=strlen(data);
			if (!ldata || *data=='\n' || *data=='\r')  /* last header line */
			    break;
			/* let's parse the request */
			if (!strncasecmp(data,"Content-length: ",16)) {
			    contlen=atoi(data+16);
			}
			if (!strncmp(data,"If-Modified-Since",17))
			    if (!NoCache)
				ifmod=1;
			    else
				continue;  /* if "nocache", forget this line */
			if (!strncmp(data,"Pragma: no-cache",16))
			    if (!NoCache)
				pragmanocache=1;
			    else
				continue;  /* if "nocache", forget this line */
			alarm(TIMEOUT);
			write(sockOut,data,ldata);
		    }
		}

		if (NoCache) {    /* if "nocache", adds a "no-cache" header */
		    alarm(TIMEOUT);
		    write(sockOut,"Pragma: no-cache\n",17);
		    pragmanocache=1;
		}
	  
		write(sockOut,"\n",1);



		/********* get post data **************/
		if (contlen) {
		    char *postdata=malloc(contlen+1);
		    alarm(TIMEOUT);
		    fgets(postdata,contlen+1,fsin);
		    alarm(TIMEOUT);
		    write(sockOut,postdata,contlen);
		    free(postdata);
		}

		/*************** retrieve data *****************/ 
		do {
		    do {
			alarm(TIMEOUT);
			ldata=read(sockOut,data,LDATA);
		    } while (ldata==-1 && errno==EINTR); /* retry on interrupt */
		    if (ldata<0)
			sayerror(ABORTOUTERR);

		    if (ldata) {            /* if ldata > 0, it's not the end yet */
			do {
			    alarm(TIMEOUT);
			    err=write(sockIn,data,ldata);
			} while (err==-1 && errno==EINTR);
			if (err==-1)
			    sayerror(ABORTINERR);
		    }
		} while (ldata>0);        /* loops while more data available */

		close(sockIn);   /* close the sockets */
		close(sockOut);
		exit(0);  /* no error */
	    }
	    default: /* the daemon itself ! (parent process)  */
		++childnbr;     /* count one more child */
		signal(SIGCHLD,sigchld);   /* ready to clear zombies */
		close(sockIn);
	    } /* end of switch(fork()) (daemon) */
	} /* end of infinite loop for(;;) (daemon) */
    default: /* the launcher */
	printf("Webroute is ready and running on port:%d with pid:%d\n",listenport,webroutepid);
	if (killtimeout)
	    printf("It will be killed after %d seconds of inactivity\n",killtimeout);
	putchar('\n');
	exit(0); /* exit normally, keeping the daemon behind us */
    } /* end of switch(fork())  (launcher) */
}  /* end of main() */
/* End of webroute.c */

